from datetime import datetime, timedelta, timezone
import tempfile
from bson.objectid import ObjectId
from .queue import Queue
from .job import Job
from .utils import executionCodeExec, executionCodeSubProcess
from typing import Dict, Type, Union, Tuple, Callable
from time import sleep
import importlib
import subprocess
import threading
import sys, json, os, traceback
from bson import json_util
from .misc import getSystemInfo

class WorkerThread(threading.Thread):
  def __init__(self, function_name, executable=None, args=[], kwargs={}, stdout=None, modulePaths=[], job=None, as_subprocess=True):
    threading.Thread.__init__(self)
    self.function_name=function_name
    self.executable=executable
    self.args=args
    self.kwargs=kwargs
    self.stdout=stdout
    self.modulePaths=modulePaths
    self.job=job
    self.output=None
    self.as_subprocess = as_subprocess

  def run(self):
    function_name=self.function_name
    executable=self.executable
    args=self.args
    kwargs=self.kwargs
    stdout=self.stdout
    modulePaths=self.modulePaths

    if executable is None:
      executable = sys.executable

    if modulePaths is None:
      modulePaths = []

    with tempfile.TemporaryDirectory() as tmp_dir:
      tmp_file_name = os.path.join(tmp_dir, "job_results.json")

      data = {"function_name": function_name, "args": args, "kwargs": kwargs, "stdout": stdout, "modules": modulePaths + [os.getcwd(),], "output": tmp_file_name}

      if self.as_subprocess:
        _code = executionCodeSubProcess(
          job_id=str(self.job.id),
          worker_id=str(self.job.worker_id),
          conn_info=self.job.get_connection_info()
        )
        subprocess.run([executable, "-c", _code], input=str.encode(json.dumps(data, default=json_util.default)))
      else:
        _code = executionCodeExec(
          job_id=str(self.job.id),
          worker_id=str(self.job.worker_id),
          conn_info=self.job.get_connection_info()
        )
        exec(_code, {"payload": data})

      if os.path.exists(tmp_file_name):
        with open(tmp_file_name, 'r') as tmp_file:
          self.output = json.load(tmp_file)
  
  def get_output(self):
    return self.output

  def get_job(self):
    return self.job

class Worker:

  def __init__(self, queues: Union[Tuple[Queue,...], Type[Queue]], channel: Union[Tuple[str,...], str]=None, heart_beat: int = 1, verbosity: str = "error", logger: Union[str, Callable] = None, executables: Dict = None, logFile: str = None, modulePaths=[], as_subprocess=True) -> None:
    
    if isinstance(queues, tuple) or isinstance(queues, list):
      self.queues = queues
    elif isinstance(queues, Queue):
      self.queues = tuple([queues])

    if isinstance(channel, tuple) or isinstance(channel, list):
      self._channels = channel
    elif isinstance(channel, str):
      self._channels = [channel]
    else:
      self._channels = None

    self._working = False
    self._heart_beat = heart_beat
    self._running = False
    self.as_subprocess = as_subprocess

    self._verbosity = verbosity
    self.setLogger(logger)

    self.logFile = None
    if not logFile is None:
      self.logFile = logFile

    self.modulePaths = modulePaths

    if not executables is None and isinstance(executables, dict):
      self.executables = executables
    else:
      self.executables = {}

    self.worker_info = getSystemInfo()

    self.worker_id = ObjectId()
    self.status_update_frequency = 10
    self.current_status_update_timeout = 10
    self.jobs_completed = 0
    self.jobs_failed = 0
    self.last_error_message = None

    self.thread = None
    
  def setLogger(self, logger: Union[str, Callable]):
    if callable(logger):
      self._logger = logger
    elif isinstance(logger, str):
      callback = str(logger)
      if str(callback).__contains__("."):
        mod_name, func_name = callback.rsplit(".", 1)
        mod = importlib.import_module(mod_name)
        func = getattr(mod, func_name)
        self._logger = func
      elif callback in globals():
        func = globals()[callback]
        self._logger = func
    else:
      self._logger = None

  def setVerbosity(self, verbosity: str):
    self.verbosity = verbosity

  def start(self):
    self._running = True
    self._run()

  def beat_heart(self):
    self.current_status_update_timeout += 1
    if self.current_status_update_timeout >= self.status_update_frequency:
      self.current_status_update_timeout = 0
      if len(self.queues) > 0:
        c_dt = datetime.now(tz=timezone.utc)
        data = {
          "status": "working" if self._working else "waiting",
          "expireAt": datetime.now(tz=timezone.utc) + timedelta(seconds=self._heart_beat + 10 + self.status_update_frequency),
          "queues": [q.channel for q in self.queues],
          "info": self.worker_info,
          "jobs_completed": self.jobs_completed,
          "jobs_failed": self.jobs_failed,
          "last_error_message": self.last_error_message,
          "is_worker": True
        }

        if not self.thread is None:
          c_job = self.thread.get_job()
          if not c_job is None:
            data["current_job_id"] = c_job.id

        coll = self.queues[0].collection
        try:  
          if self.worker_id is None:
            data["createdAt"] = c_dt
            data["updatedAt"] = c_dt
            res = coll.insert_one(data)
            self.worker_id = res.inserted_id
          else:
            data["updatedAt"] = c_dt
            res = coll.find_one_and_update({"_id": ObjectId(self.worker_id)}, {"$set": data}, projection={"_id": True})
            if res is None:
              data["createdAt"] = c_dt
              data["updatedAt"] = c_dt
              res = coll.insert_one(data)
              self.worker_id = res.inserted_id
        except Exception as e:
          print(f"Worker status update error: {e}")
          
  def stop(self):
    self._running = False

  def _work(self):
    self._working = True
    job: Type[Job] = None

    for queue in self.queues:
      if self._channels is None:
        job = queue.dequeue()
        if not job is None:
          break
      else:
        for channel in self._channels:
          job = queue.dequeue(channel)
          if not job is None:
            break
        if not job is None:
          break

    if not job is None:
      self._run_job(job)
      # self._working = False
    else:
      self._working = False

  def _on_output(self, job, output):
    payload = job.payload
    callback = payload.get("function_name")

    if not output is None:
      if "result" in output:
        job.complete(output["result"])
        self.jobs_completed += 1
      if "error" in output:
        print(output["error"], flush=True)
        job.error(output["error"])
        self.jobs_failed += 1
        self.last_error_message = output["error"] if isinstance(output["error"], dict) else dict(message=output["error"], trace="")
        self.last_error_message["errorAt"] = datetime.utcnow()
        self.last_error_message["callback"] = str(callback)
        self.last_error_message["jobId"] = job.id
  
  def _run_job(self, job: Type[Job]):
    payload = job.payload
    job.setVerbosity(self._verbosity)
    job.setLogger(self._logger)
    callback = None
    args = None
    kwargs = None
    try:
      if not payload is None:
        callback = payload.get("function_name")

        if not callback is None:
          args = payload.get("args", [])
          kwargs = payload.get("kwargs", {})
          executable = None
          channel = job.channel

          if not channel is None:
            if channel in self.executables:
              executable = self.executables[channel]
          job.worker_id = self.worker_id
          self.thread = WorkerThread(callback, args=args, kwargs=kwargs, executable=executable, stdout=self.logFile, modulePaths=self.modulePaths, job=job, as_subprocess=self.as_subprocess)
          self.thread.start()

        else:
          job.error(f"Error: no callback funtion specified for job!")
          print(f"Error: no callback funtion specified for job!")

    except Exception as ex:
      errtrace = traceback.format_exc()
      job.error({"message": f"Error: {ex}", "trace": str(errtrace)})
      self.jobs_failed += 1
      self.last_error_message = dict(message=f"Error: {ex}", trace=str(errtrace))
      self.last_error_message["errorAt"] = datetime.utcnow()
      self.last_error_message["callback"] = callback
      self.last_error_message["jobId"] = job.id
      print(f"Error: {ex}")

  def _run(self):
    while True:
      if not self._running:
        break

      if not self._working:
        self._work()

      if not self.thread is None:
        out = self.thread.get_output()
        if not out is None:
          job = self.thread.get_job()
          self._on_output(job, out)
          self.thread.join()
          self.thread = None
          self._working = False

      self.beat_heart()
      sleep(self._heart_beat)


from .worker import Worker
from .queue import Queue
import yaml
from argparse import Action

def dict_load(data):
  data = str(data).strip()
  if not data.startswith("{"):
    data = "{" + data + "}"

  data = ": ".join([s.strip() for s in str(data).split(":")])
  return yaml.load(data, Loader=yaml.FullLoader)

class UpdateAction(Action):
  def __init__(self, option_strings, dest, nargs=None, **kwargs):
    if nargs is not None:
      raise ValueError("nargs not allowed")
    super().__init__(option_strings, dest, **kwargs)

  def __call__(self, parser, namespace, values, option_string=None):
    init_value = getattr(namespace, self.dest, {})
    init_value.update(values)
    setattr(namespace, self.dest, init_value)

def param_parser(v):
  d = "{" + f"value: {v}" + "}"
  d = dict_load(d)
  return d.get("value")

def parse_args(cmd_args=None):
  import argparse
  parser = argparse.ArgumentParser(description='Mongo Queuing and Scheduling Library', add_help=True)
  subparsers = parser.add_subparsers(dest='action')

  worker = subparsers.add_parser('worker', description='Initialize a worker instance to start working on scheduled jobs')

  worker.add_argument('channels', metavar='channels', type=str, nargs='*', default=None, help='channels to be monitored by the worker')
  worker.add_argument('-u', '--conn', dest='db_conn', type=str, default="mongodb://localhost:27017",  help='mongodb connection string (default: mongodb://localhost:27017)')
  worker.add_argument('--dbname', dest='db_name', type=str, default="jobs",  help='mongodb database name (default: jobs)')
  worker.add_argument('--colname', dest='col_name', type=str, default="jobs",  help='mongodb collection name (default: jobs)')
  worker.add_argument('--consumer-id', dest='consumerId', type=str, default=None,  help='worker consumer id (default: None)')
  worker.add_argument('-l', '--lang', dest='lang', type=str, default="python",  help='worker language (default: python)')
  worker.add_argument('-m', '--modules', dest='modules', action="append", type=str, default=[], help='additional python module paths')
  worker.add_argument('-b', '--heartbeat', dest='heartbeat', type=int, default=1,  help='worker heart beat in seconds (default: 1)')
  worker.add_argument('--logger', dest='logger', type=str, default=None,  help='logger callback function (default: None)')
  worker.add_argument('--log-file', dest='log_file', type=str, default=None,  help='path to log file (default: None)')
  worker.add_argument('--no-sub-process', dest='no_sub_process', action='store_true', help='run job with exec command instead of the default subprocess command')
  worker.add_argument('--verbosity', dest='verbosity', type=str, default="error",  help='logger verbosity, options are [error, completed, progress] (default: error)')
  worker.add_argument('-e', '--executable', dest='executables', action=UpdateAction, type=dict_load, default={},  help='python executable paths for running job based on channels.')

  queue = subparsers.add_parser('queue', description='Schedule a job into the job queue')
  queue.add_argument('function_name', metavar='function_name', type=str, help='the full name of the entry point function')
  queue.add_argument('args', metavar='args', type=param_parser, nargs="*", help='positional arguments for entry point function (Arguments are passed as strings for other data types use --kwargs)')
  queue.add_argument('-u', '--conn', dest='db_conn', type=str, default="mongodb://localhost:27017",  help='mongodb connection string (default: mongodb://localhost:27017)')
  queue.add_argument('--dbname', dest='db_name', type=str, default="jobs",  help='mongodb database name (default: jobs)')
  queue.add_argument('--colname', dest='col_name', type=str, default="jobs",  help='mongodb collection name (default: jobs)')
  queue.add_argument('--consumer-id', dest='consumerId', type=str, default=None,  help='worker consumer id (default: None)')
  queue.add_argument('-l', '--lang', dest='lang', type=str, default="python",  help='worker language (default: python)')
  queue.add_argument('-c', '--channel', dest='channel', type=str, default=None, help='channel to place the job (default: None)')
  queue.add_argument('-p', '--priority', dest='priority', type=int, default=None,  help='priority of the job (default: None)')
  queue.add_argument('-j', '--job-id', dest='job_id', type=str, default=None, help='custom id for this job (default: None)')
  queue.add_argument('-k', '--kwargs', dest='kwargs', action=UpdateAction, type=dict_load, default={}, help='named arguments for the function in yaml or json format')
  queue.add_argument('--job-timeout', dest='job_timeout', type=int, default=None,  help=f"job timeout in seconds (default: None)")
  queue.add_argument('--result-ttl', dest='result_ttl', type=int, default=None,  help='number of seconds to keep the job results (default: None)')
  queue.add_argument('--ttl', dest='ttl', type=int, default=None,  help='worker heart beat in seconds (default: None)')
  queue.add_argument('--failure-ttl', dest='failure_ttl', type=int, default=None,  help=f"number of seconds to wait upon job failure (default: None)")
  queue.add_argument('--depends_on', dest='depends_on', action="append", type=str, default=None,  help='job ids of jobs which are to be completed before this job')
  queue.add_argument('--description', dest='description', type=str, default=None,  help='description for the job')
  queue.add_argument('--on-success', dest='on_success', type=str, default=None,  help='function to be invoked on successful execution of the job')
  queue.add_argument('--on-failure', dest='on_failure', type=str, default=None,  help='function to be invoked on failure of the job')
  queue.add_argument('--max-attempts', dest='max_attempts', type=int, default=None,  help='maximum number of times to attempt executing job in case of failure')

  initialize = subparsers.add_parser('init', description='Prepares database for job scheduling and worker')
  initialize.add_argument('-u', '--conn', dest='db_conn', type=str, default="mongodb://localhost:27017",  help='mongodb connection string (default: mongodb://localhost:27017)')
  initialize.add_argument('--dbname', dest='db_name', type=str, default="jobs",  help='mongodb database name (default: jobs)')
  initialize.add_argument('--colname', dest='col_name', type=str, default="jobs",  help='mongodb collection name (default: jobs)')


  args = parser.parse_args() if cmd_args is None else parser.parse_args(cmd_args)

  return args

def main(cmd_args=None):
  args = parse_args(cmd_args=cmd_args)
  action = args.action
  
  if str(action).lower() == "worker":
    run_worker(args)
  elif str(action).lower() == "queue":
    run_queue(args)
  elif str(action).lower() == "init":
    run_init(args)

def run_worker(args):
  from pymongo import MongoClient
  import os, sys
  client = MongoClient(args.db_conn)
  queues = ()
  channels = args.channels
  kwargs = dict(db_name=args.db_name, col_name=args.col_name, db_conn=args.db_conn)

  if not args.consumerId is None:
    kwargs['consumerId'] = args.consumerId

  if not args.lang is None:
    kwargs["lang"] = args.lang

  if (channels is None) or len(channels) == 0:
    queue = Queue(client, **kwargs)
    queues += (queue,)
  else:
    for channel in channels:
      kwargs['channel'] = channel
      queue = Queue(client, **kwargs)
      queues += (queue,)
  
  if len(queues) > 0:
    sys.path.append(os.getcwd())
    for module in args.modules:
      sys.path.append(os.path.abspath(module))

    modulePaths = [os.path.abspath(module) for module in args.modules]
    
    as_subprocess = not args.no_sub_process
    worker = Worker(queues, heart_beat=args.heartbeat, logger=args.logger, verbosity=args.verbosity, executables=args.executables, logFile=args.log_file, modulePaths=modulePaths, as_subprocess=as_subprocess)
    if (channels is None) or len(channels) == 0:
      print(f"**Started worker with heart beat of {args.heartbeat} second(s)", flush=True)
    else:
      print(f"**Started worker to monitor channels", channels, f"with heart beat of {args.heartbeat} second(s)", flush=True)

    worker.start()
  else:
    print("No channels were specified")

def run_queue(args):
  from pymongo import MongoClient
  import os, sys
  client = MongoClient(args.db_conn)

  kwargs = dict(db_name=args.db_name, col_name=args.col_name, db_conn=args.db_conn)
  if not args.consumerId is None:
    kwargs['consumerId'] = args.consumerId
    
  if not args.lang is None:
    kwargs["lang"] = args.lang

  queue = Queue(client, **kwargs)

  keys = ['lang', 'channel', 'priority', 'job_timeout', 'result_ttl', 'ttl', 'failure_ttl', 'depends_on', 'job_id', 'description', 'on_success', 'on_failure', 'max_attempts', 'args', 'kwargs']
  params = {}
  for k in keys:
    if hasattr(args, k):
      params[k] = getattr(args, k)

  jobId = queue.enqueue(args.function_name, **params)
  print(f"Job Successfully queued, {jobId}")
  return jobId

def run_init(args):
  from pymongo import MongoClient
  client = MongoClient(args.db_conn)
  db = client.get_database(args.db_name)
  coll = db.get_collection(args.col_name)
  coll.create_index("expireAt", expireAfterSeconds=0)
  


from pymongo import DESCENDING, ASCENDING
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo import MongoClient
from pymongo.results import InsertOneResult
from typing import Union, Callable, Optional, Type
from datetime import datetime, timedelta, timezone
from .job import Job
from .misc import parse_time_to_seconds, toOid

class Queue:
  def __init__(self, connection: Union[str, Type[Collection], Type[Database], Type[MongoClient]], consumerId: str = "default-customer-id", lang: str = "python", channel: str = "default", priority: int = 0, job_timeout: Union[str, int, None]='1h', result_ttl: Union[str,int,None]=3600*24, ttl: Union[str,int,None]=None, failure_ttl: Union[str,int]='1w', max_attempts: int=1, db_name: Optional[str]="jobs", col_name: Optional[str]="jobs", db_conn: Optional[str]="") -> None:
    
    if isinstance(connection, MongoClient):
      db = connection[db_name]
      self.collection = db[col_name]
    elif isinstance(connection, Database):
      self.collection = connection[col_name]
    elif isinstance(connection, Collection):
      self.collection = connection
    elif isinstance(connection, str):
      self.collection = MongoClient(connection)[db_name][col_name]
    else:
      print("Connection object should be one of either a mongoclient, database or collection objects, or a string representing the connection url", "\n", "Current object type is", type(connection))
      self.collection = connection

    self.lang = lang
    self.consumerId = consumerId
    self.job_timeout = parse_time_to_seconds(job_timeout) if not job_timeout is None else job_timeout
    self.max_attempts = max_attempts
    self.channel = channel
    self.result_ttl = parse_time_to_seconds(result_ttl) if not result_ttl is None else result_ttl
    self.ttl = parse_time_to_seconds(ttl) if not ttl is None else ttl
    self.failure_ttl = parse_time_to_seconds(failure_ttl) if not failure_ttl is None else failure_ttl
    self.priority = priority
    self.db_conn = db_conn
    self.db_name = db_name
    self.db_coll = col_name

  def enqueue(self, function_name: Union[Callable,str], *_args, **_kwargs) -> Optional[str]:

    if callable(function_name):
      function_name = f"{function_name.__module__}.{function_name.__name__}"
    else:
      function_name = str(function_name)

    channel = _kwargs.get("channel")
    priority = _kwargs.get("priority")
    job_timeout = _kwargs.get("job_timeout")
    result_ttl = _kwargs.get("result_ttl")
    ttl = _kwargs.get("ttl")
    failure_ttl = _kwargs.get("failure_ttl")
    depends_on = _kwargs.get("depends_on")
    job_id = _kwargs.get("job_id")
    description = _kwargs.get("description")
    on_success = _kwargs.get("on_success")
    on_failure = _kwargs.get("on_failure")
    max_attempts = _kwargs.get("max_attempts")
    args = _kwargs.get("args")
    kwargs = _kwargs.get("kwargs")
    lang = _kwargs.get("lang", self.lang)

    removable_keys = ["lang", "channel", "priority", "job_timeout", "result_ttl", "ttl", "failure_ttl", "depends_on", "job_id", "description", "on_success", "on_failure", "max_attempts", "args", "kwargs"]
    for k in removable_keys:
      if k in _kwargs:
        del _kwargs[k]

    if result_ttl is None:
      result_ttl = self.result_ttl

    if ttl is None:
      ttl = self.ttl

    if failure_ttl is None:
      failure_ttl = self.failure_ttl

    if max_attempts is None:
      max_attempts = self.max_attempts

    if channel is None:
      channel = self.channel

    if priority is None:
      priority = self.priority

    if isinstance(kwargs, dict):
      _kwargs.update(kwargs)

    if isinstance(args, list) or isinstance(args, tuple):
      _args += tuple(args)

    if isinstance(depends_on, list) or isinstance(depends_on, tuple):
      depends_on = list(depends_on)
      for i in range(len(depends_on)):
        depends_on[i] = toOid(depends_on[i])
    elif not depends_on is None:
      depends_on = [toOid(depends_on)]

    data = {
      "data": {
        "function_name": function_name,
        "job_timeout": job_timeout,
        "description": description,
        "on_success": on_success,
        "on_failure": on_failure,
        "args": _args,
        "kwargs": _kwargs,
      },
      "item_type": "queue",
      "consumer_id": self.consumerId,
      "depends_on": depends_on,
      "result_ttl": result_ttl,
      "ttl": ttl,
      "failure_ttl": failure_ttl,
      "max_attempts": max_attempts,
      "channel": channel,
      "inProgress": False,
      "done": False,
      "attempts": 0,
      "progress": 0,
      "priority": priority,
      "lang": lang,
      "createdAt": datetime.now(timezone.utc),
    }

    if not (ttl is None):
      if int(ttl) > 0:
        data["expireAt"] = datetime.now(timezone.utc) + timedelta(seconds=int(ttl))

    if not job_id is None:
      data["_id"] = toOid(job_id)

    res: Type[InsertOneResult]  = self.collection.insert_one(data)
    return res.inserted_id  

  def dequeue(self, channel: Optional[str]=None, lang: str = None, job_id: Optional[str]=None) -> Optional[Type[Job]]:
    res = self.collection.find({"done": True}, ["_id"])
    ids = []

    if lang is None:
      lang = self.lang

    for item in list(res):
      oid = item.get("_id")
      if not oid is None:
        ids.append(toOid(oid))

    query = {"lang": lang, "consumer_id": self.consumerId, "item_type": "queue", "inProgress": False, "done": False, "$expr": {"$lt": ["$attempts", "$max_attempts"]}, "$or": [{"depends_on": None}, {"depends_on": {"$not": {"$elemMatch": {"$nin" : ids }}}}]}
    if channel is None:
      channel = self.channel

    query["channel"] = channel
    if not job_id is None:
      query["_id"] = toOid(job_id)

    job = self.collection.find_one_and_update(query, {"$set": {"inProgress": True, "startedAt": datetime.now(timezone.utc)}}, sort=[("priority", DESCENDING), ("createdAt", ASCENDING)])

    if not job is None:
      keys = ["depends_on", "result_ttl", "failure_ttl", "ttl", "createdAt", "lang", "channel"]
      kwargs = {}
      for k in keys:
        if k in job:
          kwargs[k] = job.get(k)

      conn_info = dict(db_conn=self.db_conn, db_name=self.db_name, db_coll=self.db_coll)
      return Job(job.get("_id"), job.get("data", {}), self.collection, conn_info, **kwargs)

    return None


from pymongo.collection import Collection
from typing import Type, Callable
from datetime import datetime, timedelta
from .misc import toOid
import importlib

class Job:
  def __init__(self, id, data, collection: Type[Collection], connection_info=None, **kwargs):
    self.id = toOid(id)
    self.coll = collection
    self.payload = data
    self.kwargs = kwargs
    self.result = None
    self.logger = None
    self.channel = None
    self.worker_id = None
    self.verbosity = "error"
    self.__connection_info = connection_info

    if "logger" in kwargs:
      self.setLogger(kwargs["logger"])

    if "verbosity" in kwargs:
      self.setVerbosity(kwargs["verbosity"])

    if "channel" in kwargs:
      self.channel = kwargs["channel"]

  def data(self):
    return self.payload

  def set_result(self, result):
    self.result = result
    self.coll.update_one({"_id": self.id}, {"$set": {"result": self.result}})
    return True

  def complete(self, result=None):
    if result is None:
      result = self.result
    
    set_data = {"progress": 100, "inProgress": False, "done": True, "result": result, "completedAt": datetime.utcnow()}
    if "result_ttl" in self.kwargs:
      ttl = self.kwargs.get("result_ttl")
      if int(ttl) > 0:
        set_data["expireAt"] = datetime.utcnow() + timedelta(seconds=int(ttl))

    self.coll.update_one({"_id": self.id}, {"$set": set_data})

    
    if "on_success" in self.payload:
      payload = self.payload
      try:
        if not payload is None:
          callback = payload.get("on_success")
          if not callback is None:
            if str(callback).__contains__("."):
              mod_name, func_name = callback.rsplit(".", 1)
              mod = importlib.import_module(mod_name)
              func = getattr(mod, func_name)
              args = payload.get("args", [])
              kwargs = payload.get("kwargs", {})
              func(result, *args, **kwargs)
            elif callback in globals():
              func = globals()[callback]
              args = payload.get("args", [])
              kwargs = payload.get("kwargs", {})
              func(result, *args, **kwargs)

      except Exception as ex:
        self.log("error", error_message=f"Error: {ex}")

    self.log("completed", result=result)
    return True

  def error(self, msg=None):   
    set_data = {"inProgress": False, "error": True, "lastErrorAt": datetime.utcnow(), "errorMessage": msg}

    if "failure_ttl" in self.kwargs:
      ttl = self.kwargs.get("failure_ttl")
      if int(ttl) > 0:
        set_data["expireAt"] = datetime.utcnow() + timedelta(seconds=int(ttl))
    
    self.coll.update_one({"_id": self.id}, {"$inc": {"attempts": 1}, "$set": set_data})
    
    if "on_failure" in self.payload:
      payload = self.payload
      try:
        if not payload is None:
          callback = payload.get("on_failure")
          if not callback is None:
            if str(callback).__contains__("."):
              mod_name, func_name = callback.rsplit(".", 1)
              mod = importlib.import_module(mod_name)
              func = getattr(mod, func_name)
              args = payload.get("args", [])
              kwargs = payload.get("kwargs", {})
              func(msg, *args, **kwargs)
            elif callback in globals():
              func = globals()[callback]
              args = payload.get("args", [])
              kwargs = payload.get("kwargs", {})
              func(msg, *args, **kwargs)

      except Exception as ex:
        self.log("error", error_message=f"Error: {ex}")

    self.log("error", error_message=msg)
    return True

  def log(self, type, *args, **kwargs):
    if not self.logger is None:
      if self.verbosity.lower().find(type) >= 0:
        payload = self.payload
        _args = payload.get("args", [])
        _kwargs = payload.get("kwargs", {})
        _args = _args + list(args)
        _kwargs.update(kwargs)
        self.logger(type, *_args, **_kwargs)

  def progress(self, percent, msg=None):
    self.coll.update_one({"_id": self.id}, {"$set": {"progress": percent, "progressMessage": msg, "lastProgressAt": datetime.utcnow()}})
    self.log("progress", percent=percent, progress_message=msg)
    return True

  def setLogger(self, logger):
    self.logger = logger

  def setVerbosity(self, verbosity):
    self.verbosity = verbosity

  def release(self):
    self.coll.update_one({"_id": self.id}, {"$set": {"inProgress": False, "error": False, "done": False, "releasedAt": datetime.utcnow(), "attempts": 0}})
    return True
  
  def set_connection_info(self, info):
    self.__connection_info = info

  def get_connection_info(self):
    return self.__connection_info



def executionCodeExec(job_id, worker_id, conn_info):
  code = f"""
import sys, json, os, importlib, traceback
from bson import json_util

try:
  from mqas import worker_utils
  worker_utils.set_value("__job_id", "{job_id}")
  worker_utils.set_value("__worker_id", "{worker_id}")
  worker_utils.set_value("__db_conn", "{conn_info.get('db_conn')}")
  worker_utils.set_value("__db_name", "{conn_info.get('db_name')}")
  worker_utils.set_value("__db_coll", "{conn_info.get('db_coll')}")
except:
  pass

  """
  code += """
def write_output(data, filename):
  if not filename is None:
    with open(filename, "w") as json_file:
      json.dump(data, json_file, default=json_util.default)

try:
  modules = payload.get("modules", [])
  
  for module in modules:
    sys.path.append(os.path.abspath(module))

  output_file = payload.get("output")

  if not payload is None:
    callback = payload.get("function_name")
    if not callback is None:
      if str(callback).__contains__("."):
        mod_name, func_name = callback.rsplit(".", 1)
        mod = importlib.import_module(mod_name)
        func = getattr(mod, func_name)
        args = payload.get("args", [])
        kwargs = payload.get("kwargs", {})
        result = func(*args, **kwargs)
        data = {"result": result}
        write_output(data, output_file)
      elif callback in globals():
        func = globals()[callback]
        args = payload.get("args", [])
        kwargs = payload.get("kwargs", {})
        result = func(*args, **kwargs)
        data = {"result": result}
        write_output(data, output_file)
      else:
        err = "Function " + str(callback) + " not found!"
        raise Exception(err)

except Exception as ex:
  errtrace = traceback.format_exc()
  err = "Error " + str(ex)
  data = {"error": {"trace": str(errtrace), "message": err}}
  write_output(data, output_file)

  """
  return code

def executionCodeSubProcess(job_id, worker_id, conn_info):
  code = f"""
import sys, json, os, importlib, traceback
from bson import json_util

try:
  from mqas import worker_utils
  worker_utils.set_value("__job_id", "{job_id}")
  worker_utils.set_value("__worker_id", "{worker_id}")
  worker_utils.set_value("__db_conn", "{conn_info.get('db_conn')}")
  worker_utils.set_value("__db_name", "{conn_info.get('db_name')}")
  worker_utils.set_value("__db_coll", "{conn_info.get('db_coll')}")
except:
  pass

  """
  code += """
def write_output(data, filename):
  if not filename is None:
    with open(filename, "w") as json_file:
      json.dump(data, json_file, default=json_util.default)

try:
  payload = json.loads(sys.stdin.read(), object_hook=json_util.object_hook)
  modules = payload.get("modules", [])
  
  for module in modules:
    sys.path.append(os.path.abspath(module))

  output_file = payload.get("output")

  if not payload is None:
    callback = payload.get("function_name")
    if not callback is None:
      if str(callback).__contains__("."):
        mod_name, func_name = callback.rsplit(".", 1)
        mod = importlib.import_module(mod_name)
        func = getattr(mod, func_name)
        args = payload.get("args", [])
        kwargs = payload.get("kwargs", {})
        result = func(*args, **kwargs)
        data = {"result": result}
        write_output(data, output_file)
      elif callback in globals():
        func = globals()[callback]
        args = payload.get("args", [])
        kwargs = payload.get("kwargs", {})
        result = func(*args, **kwargs)
        data = {"result": result}
        write_output(data, output_file)
      else:
        err = "Function " + str(callback) + " not found!"
        raise Exception(err)

except Exception as ex:
  errtrace = traceback.format_exc()
  err = "Error " + str(ex)
  data = {"error": {"trace": str(errtrace), "message": err}}
  write_output(data, output_file)

  """
  return code


import re
from datetime import timedelta
from bson.objectid import ObjectId
import platform,socket,re,uuid,psutil,logging


regex = re.compile(r'((?P<weeks>\d+?)w)?((?P<days>\d+?)d)?((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?')


def parse_time(time_str):
    parts = regex.match(time_str)
    if not parts:
      return
    parts = parts.groupdict()
    time_params = {}
    for name, param in parts.items():
      if param:
        time_params[name] = int(param)

    return timedelta(**time_params)

def parse_time_to_seconds(time_str):
    if isinstance(time_str, int) or isinstance(time_str, float):
      return int(time_str)
      
    parts = regex.match(time_str)
    if not parts:
      return
    parts = parts.groupdict()
    time_params = {}
    for name, param in parts.items():
      if param:
        time_params[name] = int(param)

    return timedelta(**time_params).total_seconds()

def toOid(id):
  oid = id
  try:
    oid = ObjectId(oid)
  except:
    pass
  
  return oid

def getSystemInfo():
  try:
    info={}
    info['platform']=platform.system()
    info['platform-release']=platform.release()
    info['platform-version']=platform.version()
    info['architecture']=platform.machine()
    info['hostname']=socket.gethostname()
    info['ip-address']=socket.gethostbyname(socket.gethostname())
    info['mac-address']=':'.join(re.findall('..', '%012x' % uuid.getnode()))
    info['processor']=platform.processor()
    info['ram']=str(round(psutil.virtual_memory().total / (1024.0 **3)))+" GB"
    return info
  except Exception as e:
    logging.exception(e)
    return {}


import threading, psutil, datetime, time
from pymongo.collection import Collection
from pymongo import MongoClient
from bson.objectid import ObjectId as _ObjectId

__STORAGE = threading.local()

def make_oid(oid=None):
    """A function to convert string to mongodb ObjectId or generate a new ObjectId

    Parameters
    ----------
    oid : str, Optional
      A string to be converted to ObjectId

    Returns
    -------
    objectId: ObjectId
      The ObjectId representation of oid or new ObjectId

    """
    try:
        if oid is None:
            return _ObjectId()

        if isinstance(oid, (list, tuple)):
            return list((_ObjectId(oi) for oi in oid))
        
        return _ObjectId(oid)
    except Exception:
        return None

def set_value(key, value):
    setattr(__STORAGE, key, value)

def get_value(key, default=None):
    if hasattr(__STORAGE, key):
        return getattr(__STORAGE, key)
    else:
        return default

def get_job_id():
    job_id = get_value("__job_id")
    return job_id

def get_worker_id():
    worker_id = get_value("__worker_id")
    return worker_id

def get_jobs_db():
    DB_CLIENT = get_value("__db_client")
    DB_OBJ = get_value("__db_obj")

    if DB_CLIENT is None:
        db_name = get_value("__db_name")
        db_conn = get_value("__db_conn")

        if db_conn is None or db_name is None:
            return None
        
        DB_CLIENT = MongoClient(db_conn)
        DB_OBJ = DB_CLIENT[db_name]

    return DB_OBJ

def get_jobs_coll() -> Collection:
    db = get_jobs_db()
    db_coll = get_value("__db_coll")

    if db is None or db_coll is None:
        return None
    
    return db[db_coll]


class MemoryLocker:
    def __init__(self, memory_mb=None, timeout=60, update_interval=10):
        self.memory = memory_mb
        self.reservation_id = None
        self.__timeout = timeout
        self.__update_interval = min(5, update_interval, timeout)
        self.__updater_thread = None
        self.__stop_event = threading.Event()
        self.coll = get_jobs_coll()

    def lock(self, memory_mb):
        self.memory = memory_mb
        return self.__wait()

    def release(self):
        self.__stop()

    def __request_scheduling(self, worker_id):
        expire_at = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(seconds=5)
        result = self.coll.find_one_and_update(
            {
                "is_reservation_lock": True,
                "is_locked": False
            },
            {
                "$setOnInsert": {
                    "is_reservation_lock": True,
                    "is_locked": True,
                    "worker_id": worker_id,
                    "expireAt": expire_at
                }
            },
            upsert=True,
            return_document=True
        )

        if result:
            return result.get("_id")

        return None

    def __reserve_memory(self):
        worker_id = get_worker_id()

        lock_id = self.__request_scheduling(worker_id)

        if lock_id is None:
            return False
        
        avail_mem = psutil.virtual_memory().available / (1024**2) # MB
        
        job_id = get_job_id()
        # Get reserved memory
        coll = self.coll
        
        workers = coll.find({ "is_worker": True }, ["_id"])
        worker_ids = list(w.get("_id") for w in list(workers) if str(w.get("_id")) != worker_id)

        query = {
            "is_reservation": True,
            "worker_id": { "$in": worker_ids }
        }

        res = coll.find(query, ["ram"])
        total_res = sum(r.get("ram", 0) for r in res)
        can_reserve = (avail_mem - total_res) >= self.memory

        if can_reserve:
            res = coll.find_one({"worker_id": make_oid(worker_id), "is_reservation": True}, ["_id"])
            if res:
                self.reservation_id = res.get("_id")
                coll.find_one_and_update({"_id": make_oid(self.reservation_id)}, {"$set": {"ram": self.memory, "job_id": make_oid(job_id)}}, projection={"_id": True})
            else:
                data = {
                    "is_reservation": True,
                    "worker_id": make_oid(worker_id),
                    "job_id": make_oid(job_id),
                    "ram": self.memory,
                    "expireAt": datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=self.__timeout)
                }
                res = coll.insert_one(data)
                self.reservation_id = res.inserted_id
            
            coll.delete_one({"_id": lock_id})

            return True
        
        coll.delete_one({"_id": lock_id})

        return False

    def __start_updater(self):
        """Background thread that refreshes the reservation periodically."""
        def updater():
            coll = self.coll
            while not self.__stop_event.is_set() and self.reservation_id is not None:
                try:
                    coll.find_one_and_update(
                        {"_id": make_oid(self.reservation_id)},
                        {"$set": {"expireAt": datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=self.__timeout)}}
                    )
                except Exception as e:
                    print("MemoryLocker updater error:", e, flush=True)

                # Wait for update interval or stop event
                self.__stop_event.wait(self.__update_interval)

        self.__updater_thread = threading.Thread(target=updater, daemon=True)
        self.__updater_thread.start()

    def __wait(self):
        if self.coll is None:
            return
        
        while not self.__reserve_memory():
            time.sleep(1)

        self.__start_updater()

    def __enter__(self):
        if not self.memory:
            return self
        
        self.__wait()
        
        return self
    
    def __stop(self):
        # Stop the updater thread
        self.__stop_event.set()
        if self.__updater_thread:
            self.__updater_thread.join(timeout=1)

        # Release reservation
        if self.reservation_id:
            coll = self.coll
            coll.delete_one({"_id": make_oid(self.reservation_id)})
            self.reservation_id = None
        
    def __exit__(self, exc_type, exc_value, traceback):
        return self.__stop()